{-| Author: Segev Elazar Mittelman, 2021
    Adopted from: Generating Constrained Random Data with Uniform Distribution
              by  Koen Claessen, Jonas Duregaard, and Michal Palka.
    For: CMSC 838g - Fall 2021. |-} 

{-# LANGUAGE GADTs #-}

module Main where 
import Data.Maybe
import System.Environment

data Nat = Z | Suc Nat 

natToInteger :: Nat -> Integer
natToInteger Z = 0
natToInteger (Suc n) = 1 + natToInteger n

instance Show Nat where
    show = show . natToInteger

instance Eq Nat where
    Z == Z = True
    (Suc n) == (Suc m) = n == m
    _ == _ = False


instance Ord Nat where
    Z <= _ = True
    _ <= Z = False
    (Suc n) <= (Suc m) = n <= m

data ListNat = Nill | Cons Nat ListNat 

listNatToIntegerList :: ListNat -> [Integer]
listNatToIntegerList Nill = []
listNatToIntegerList (Cons n ns) = natToInteger n : listNatToIntegerList ns 

instance Show ListNat where
    show = show . listNatToIntegerList

data Tree = Leaf | Node Nat Tree Tree
    deriving Show

data Space a where
    Empty :: Space a
    Pure :: a -> Space a
    (:+:) :: Space a -> Space a -> Space a
    (:*:) :: Space a1 -> Space a2 -> Space (a1,a2)
    Pay :: Space a -> Space a
    (:$:) :: (a1 -> a2) -> Space a1 -> Space a2

(<*>) :: Space (a -> b) -> Space a -> Space b
s1 <*> s2 = (\(f, a) -> (f a)) :$: (s1 :*: s2)

spaceNat :: Space Nat
spaceNat = Pay (Pure Z :+: (Suc :$: spaceNat))

spaceListNat :: Space ListNat
spaceListNat = Pay (Pure Nill :+: (Cons :$: spaceNat Main.<*> spaceListNat))

spaceTree :: Space Tree
spaceTree = Pay (Pure Leaf :+: (Node :$: spaceNat Main.<*> spaceTree Main.<*> spaceTree))

data FinSet a where
    EmptySet :: FinSet a
    Single :: a -> FinSet a
    Product :: FinSet a1 -> FinSet a2 -> FinSet (a1, a2)
    Union :: FinSet a -> FinSet a -> FinSet a
    Apply :: (a1 -> a2) -> FinSet a1 -> FinSet a2

example1 :: FinSet (Nat, Char)
example1 = Product 
            (Apply Suc (Union (Single Z) (Union (Single (Suc Z)) (Single (Suc (Suc Z)))))) 
            (Union (Single 'A') (Union (Single 'B') EmptySet))

finSetCardinality :: FinSet a -> Integer
finSetCardinality EmptySet = 0
finSetCardinality (Single _) = 1
finSetCardinality (Product finSetA finSetB) = 
    finSetCardinality finSetA * finSetCardinality finSetB
finSetCardinality (Union finSetA1 finSetA2) = 
    finSetCardinality finSetA1 + finSetCardinality finSetA2
finSetCardinality (Apply _ finSet) = finSetCardinality finSet

indexFin :: FinSet a -> Integer -> Maybe a
indexFin EmptySet _ = Nothing
indexFin (Single a) 0 = Just a
indexFin (Single _) _ = Nothing
indexFin (Union fsa _) i | i < finSetCardinality fsa = indexFin fsa i 
indexFin (Union fsa fsb) i = indexFin fsb (i - finSetCardinality fsa)
indexFin (Product fsa fsb) i = do
    fst <- indexFin fsa (i `div` finSetCardinality fsb)
    snd <- indexFin fsb (i `mod` finSetCardinality fsb)  
    return (fst, snd)                                    
indexFin (Apply f finSet) i = do
    val <- indexFin finSet i
    return (f val) 


isS :: Nat -> Bool
isS Z = False
isS (Suc _) = True

isSValid :: Bool
isSValid = isS (Suc undefined) == True


--index :: (a -> Bool) -> Space a -> Int -> Integer -> Space a
--index predicate (constructor :$: space') size index = case valid predicate' of
--    Just _ -> constructor :$: space'
--    Nothing -> constructor :$: index predicate' space' size index
--    where predicate' = predicate . constructor


--index isS (Suc :$: spaceNat) 2 0 == ... == 
--    index (isS . Suc) (Suc :$: spaceNat) 1 0 == 
--        Suc :$: spaceNat  

--valid :: (a -> Bool) -> Maybe Bool
--valid p | crashes (p undefined) = Nothing
--valid P = Some (p undefined)

--valid isS == Nothing --needs to inspect input so fails
--valid (isS . S) == Just True --(isS . S) n == True for all naturals n.


sized :: Space a -> Integer -> FinSet a
sized Empty _ = EmptySet
sized (Pure a) 0 = Single a
sized (Pure _) _ = EmptySet
sized (Pay _) 0 = EmptySet
sized (Pay a) k = sized a (k - 1)
sized (a :+: b) k = Union (sized a k) (sized b k)
sized (f :$: a) k = Apply f (sized a k)
sized (a :*: b) k = sizedHelper (a :*: b) 0 k
    where
        sizedHelper :: Space a -> Integer -> Integer -> FinSet a
        sizedHelper (a :*: b) k1 k | k1 <= k = 
            Union (Product (sized a k1) (sized b (k - k1))) (sizedHelper (a :*: b) (k1 + 1) k)
        sizedHelper _ _ _ = EmptySet

indexSized :: Space a -> Integer -> Integer -> Maybe a
indexSized space size index = indexFin (sized space size) index 

enumSpaceSized :: Space a -> Integer -> [a]
enumSpaceSized s k = map fromJust (takeWhile isJust (map (indexSized s k) [0..]))

enumFilteredSpaceSized :: (a -> Bool) -> Space a -> Integer -> [a]
enumFilteredSpaceSized p s k = filter p $ enumSpaceSized s k

failsSorted :: ListNat -> Bool
failsSorted (Cons (Suc _) (Cons Z _)) = True
failsSorted _ = False

failsAscending :: ListNat -> Bool
failsAscending (Cons _ (Cons Z _)) = True
failsAscending _ = False

sortedListNat :: ListNat -> Bool
sortedListNat Nill = True
sortedListNat (Cons _ Nill) = True
sortedListNat (Cons x (Cons y xs)) = x <= y && sortedListNat (Cons y xs)

demo1 :: IO ()
demo1 = do
    size : _ <- getArgs
    print $ enumSpaceSized spaceTree (read size)

demo2 :: IO ()
demo2 = do
    size : _ <- getArgs  
    let allOfSize = enumSpaceSized spaceListNat (read size)
    let failingClass = enumFilteredSpaceSized failsSorted spaceListNat (read size)
    print allOfSize
    print (length allOfSize)
    print failingClass
    print (length failingClass)
    print $ fromIntegral (length failingClass) / fromIntegral (length allOfSize)




main :: IO ()
main = demo2

    
